/* Define some constants for access on chip resources */
unsigned long int* const PIO_PERC   = (unsigned long int* const) 0xfffff800;
unsigned long int* const PIO_OERC   = (unsigned long int* const) 0xfffff810;
unsigned long int* const PIO_ODRC   = (unsigned long int* const) 0xfffff814;
unsigned long int* const PIO_SODRC  = (unsigned long int* const) 0xfffff830;
unsigned long int* const PIO_CODRC  = (unsigned long int* const) 0xfffff834;
unsigned long int* const PIO_PDSRC  = (unsigned long int* const) 0xfffff83c;

/* ADC registers */
unsigned long int* const ADC_CR     = (unsigned long int* const) 0xfffd8000;
unsigned long int* const ADC_MR     = (unsigned long int* const) 0xfffd8004;
unsigned long int* const ADC_CHER   = (unsigned long int* const) 0xfffd8010;
unsigned long int* const ADC_SR     = (unsigned long int* const) 0xfffd801c;
unsigned long int* const ADC_CDR0   = (unsigned long int* const) 0xfffd8030;
unsigned long int* const ADC_CDR1   = (unsigned long int* const) 0xfffd8034;
unsigned long int* const ADC_CDR2   = (unsigned long int* const) 0xfffd8038;
unsigned long int* const ADC_CDR3   = (unsigned long int* const) 0xfffd803c;

/* Brownout detector related stuff */
unsigned long int* const SUPC_BOMR = (unsigned long int* const) 0xfffffd14;
unsigned long int* const SUPC_MR = (unsigned long int* const) 0xfffffd18;
unsigned long int* const SUPC_SR = (unsigned long int* const) 0xfffffd24;

/* Font table for digits 0-9 and A-F */
/* The segment mapping must match the big LCD lookup table below */
/* Yes, it's an ugly hack, but it does work. */
/* Feel free to do a better one! */
const unsigned char seven_segment_table[]=
{
   0x7b,    /* 0 */
   0x48,    /* 1 */
   0x3d,    /* 2 */
   0x6d,    /* 3 */
   0x4e,    /* 4 */
   0x67,    /* 5 */
   0x77,    /* 6 */
   0x49,    /* 7 */
   0x7f,    /* 8 */
   0x6f,    /* 9 */
   0x5f,    /* A */
   0x76,    /* B */
   0x34,    /* C */
   0x7c,    /* D */
   0x37,    /* E */
   0x17,    /* F */
};

struct lcd_digit_struct
{
   unsigned char address;
   unsigned char bitmask;
};

/* Table of LCD segments 
   Eight entries for each of the 12 main digits (left to right) 
   This table includes segments a-h, but not i, the comma 
   Again, this is an ugly hack and needs to be cleaned up 
   It's an improvement over the last one, but I think I can
   get it down to half this size.
   Appended the exponent digits to the end of the table.
   The exponents don't have decimal points, so I picked an unmapped
   bit so that it wouldn't break the logic.  0x50, 0x20 is valid
   memory, but it doesn't turn on a segment. */
const struct lcd_digit_struct lcd_digit_table[]=
{
   {0x10, 0x7},
   {0x10, 0x6},
   {0x18, 0x7},
   {0x10, 0x8},
   {0x18, 0x6},
   {0x20, 0x6},
   {0x20, 0x7},
   {0x28, 0x8},

   {0x10, 0x9},
   {0x18, 0x8},
   {0x18, 0x9},
   {0x18, 0xa},
   {0x20, 0x8},
   {0x20, 0x9},
   {0x20, 0xa},
   {0x28, 0xa},

   {0x10, 0xb},
   {0x10, 0xa},
   {0x18, 0xb},
   {0x10, 0xc},
   {0x20, 0xb},
   {0x28, 0xb},
   {0x18, 0xc},
   {0x20, 0xc},

   {0x10, 0xe},
   {0x10, 0xd},
   {0x18, 0xd},
   {0x18, 0xe},
   {0x20, 0xd},
   {0x28, 0xd},
   {0x20, 0xe},
   {0x20, 0xf},

   {0x10, 0x10},
   {0x10, 0xf},
   {0x18, 0x10},
   {0x10, 0x11},
   {0x18, 0xf},
   {0x28, 0xf},
   {0x20, 0x10},
   {0x28, 0x11},

   {0x10, 0x12},
   {0x18, 0x11},
   {0x18, 0x12},
   {0x18, 0x13},
   {0x20, 0x11},
   {0x20, 0x12},
   {0x20, 0x13},
   {0x28, 0x13},

   {0x10, 0x14},
   {0x10, 0x13},
   {0x18, 0x14},
   {0x10, 0x15},
   {0x20, 0x14},
   {0x28, 0x14},
   {0x18, 0x15},
   {0x20, 0x15},

   {0x10, 0x17},
   {0x10, 0x16},
   {0x18, 0x16},
   {0x18, 0x17},
   {0x20, 0x16},
   {0x28, 0x16},
   {0x20, 0x17},
   {0x20, 0x18},

   {0x10, 0x19},
   {0x10, 0x18},
   {0x18, 0x19},
   {0x10, 0x1a},
   {0x18, 0x18},
   {0x28, 0x18},
   {0x20, 0x19},
   {0x28, 0x1a},

   {0x10, 0x1b},
   {0x18, 0x1a},
   {0x18, 0x1b},
   {0x18, 0x1c},
   {0x20, 0x1a},
   {0x20, 0x1b},
   {0x20, 0x1c},
   {0x28, 0x1c},

   {0x10, 0x1d},
   {0x10, 0x1c},
   {0x18, 0x1d},
   {0x10, 0x1e},
   {0x20, 0x1d},
   {0x28, 0x1d},
   {0x18, 0x1e},
   {0x20, 0x1e},

   {0x14, 0x0},
   {0x10, 0x1f},
   {0x18, 0x1f},
   {0x1c, 0x0},
   {0x20, 0x1f},
   {0x28, 0x1f},
   {0x24, 0x0},
   {0x2c, 0x1},

   {0x14, 0x2},
   {0x14, 0x1},
   {0x24, 0x2},
   {0x1c, 0x2},
   {0x24, 0x1},
   {0x2c, 0x2},
   {0x2c, 0x3},
   {0x50, 0x20},

   {0x14, 0x4},
   {0x14, 0x3},
   {0x1c, 0x3},
   {0x1c, 0x4},
   {0x24, 0x3},
   {0x2c, 0x4},
   {0x24, 0x4},
   {0x50, 0x20},

   {0x1c, 0x6},
   {0x14, 0x5},
   {0x1c, 0x5},
   {0x24, 0x6},
   {0x24, 0x5},
   {0x2c, 0x5},
   {0x2c, 0x6},
   {0x50, 0x20},
};


/* Place a hex digit (value) at position (digit) on the main 12 digit display */
/* Values > 15 will show as a blank */
void write_char(int digit, int value)
{
   int pattern;
   int bit;
   unsigned long int address;

   /* Font lookup--convert to segments */
   pattern=seven_segment_table[value];

   /* Blank if out of range */
   if (value>15)
      pattern=0;

   /* Look up each segment in the main LCD table and turn 'em on or off */
   for (bit=0; bit<8; bit++)
   {
      /* Grab address of the segment */
      address=0xfffb4200+lcd_digit_table[(digit*8)+bit].address;

      /* Does font call for segment to be on? */
      if (pattern & (1 << bit))
      {
         /* Yep, so look up the bit and set it */
         *(unsigned long int*)address|=(1 << lcd_digit_table[(digit*8)+bit].bitmask);
      }
      else
      {
         /* Segment should be off; lookup and turn if off */
         *(unsigned long int*)address&=~(1 << lcd_digit_table[(digit*8)+bit].bitmask);
      }
   }
}

/* Turn off all the LCD segments */
void clear_screen()
{
   unsigned long int* i;
   for (i=(unsigned long int*)0xfffb4200; i<(unsigned long int*)0xfffb4250; i++)
   {
      *i=0;
   }
}

/* Reset and configure the ADC.  No idea if 
   the timing is close to correct, but it works
   well enough to map the pins on J7 and
   determine that 1.8V is full scale. */
void adc_init()
{
   /* Reset adc */
   *ADC_CR = 0x01;

   /* Configure clocking (guessing, haven't checked the math) */
   *ADC_MR=0x04041000;

   /* Enable all four channels */
   *ADC_CHER=0xf;

   /* Ready to start conversions */
}

/* Display four 12 bit hex values on the display.
   Had to use the exponent area, too. */
void show_adc(int ch0, int ch1, int ch2, int ch3)
{
   /* ch0 at far left */
   write_char(0, (ch0 >> 8) & 0x0f);
   write_char(1, (ch0 >> 4) & 0x0f);
   write_char(2, (ch0 >> 0) & 0x0f);

   write_char(4, (ch1 >> 8) & 0x0f);
   write_char(5, (ch1 >> 4) & 0x0f);
   write_char(6, (ch1 >> 0) & 0x0f);

   /* ch2 at far right */
   write_char(8, (ch2 >> 8) & 0x0f);
   write_char(9, (ch2 >> 4) & 0x0f);
   write_char(10, (ch2 >> 0) & 0x0f);

   /* ch3 in exponent area */
   write_char(12, (ch3 >> 8) & 0x0f);
   write_char(13, (ch3 >> 4) & 0x0f);
   write_char(14, (ch3 >> 0) & 0x0f);
}

int main(void)
{
   int ch0=0;
   int ch1=0;
   int ch2=0;
   int ch3=0;

   /* Configure ADC */
   adc_init();

   /* Clear screen */
   clear_screen();

   for (;;)
   {
      /* Start conversions */
      *ADC_CR = 0x02;

      /* Spin until all four channels have been converted */
      while ((*ADC_SR) & 0x0f)
      {
         if ((*ADC_SR) & 0x01)
         {
            /* Channel 0 ready, read it */
            ch0=*ADC_CDR0;
         }

         if ((*ADC_SR) & 0x03)
         {
            // channel 0 ready
            ch1=*ADC_CDR1;
         }

         if ((*ADC_SR) & 0x04)
         {
            // channel 0 ready
            ch2=*ADC_CDR2;
         }

         if ((*ADC_SR) & 0x08)
         {
            // channel 0 ready
            ch3=*ADC_CDR3;
         }

         /* Update display */
         show_adc(ch0, ch1, ch2, ch3);
      }
   }
   return 0;
}
